-- Skript zum Pseudonymisieren der persoenlichen Daten der Studierenden des flexnow-Schemas einer Datenbank-Instanz von FlexNow.
-- Ist vom 05.07.2019 (die vorige Version war vom 27.09.2018), von Peter Fasching, Ruhr-Uni Bochum, peter.fasching@rub.de
-- Die Zeilenende-Zeichen in dieser Datei sind fuer Unix-/Linux-Systeme, da in Bochum dieses Skript auf so einem System ausgefuehrt wird.
--
-- Die Spalten bez und eMail der Tabelle fn2meta.PersonStub muessten auch noch pseudonymisiert werden. Allerdings sind in der Tabelle
-- auch Pruefer/Stellvertreter enthalten, und die wollen wir nicht pseudonymisieren. Zudem steht in der bez der Name bei einem Teil
-- der Personen als "[Vorname], [Nachname]", bei vielen Personen (mit hoher Person-ID) hingegen ohne Komma, so dass ich bei Namen
-- mit mehr als einem Leerzeichen nicht mehr sicher zwischen Vor- und Nachnamen unterscheiden kann. Eventuell kuemmere ich mich spaeter noch darum.
--
-- Voraussetzungen zum Verwenden des Skriptes:
-- a) Dies ist ein SQL-Skript fuer Oracle, das auch als PL/SQL-Skript lauffaehig ist und daher regelmaessig ausfuehrbar ist.
--    Das Skript funktioniert vermutlich zumindest ab Version 11, aber sicher ab Version 12c.
-- b) Die Statements setzen voraus, dass es in der Tabelle STUDENT die Spalte testPerson gibt. Sie wurde durch das SQL-Statement fuer
--    fn2Module mit der Nummer 498 vom 29.11.2018 eingefuegt. Dieses Spalte und deren Verwendung im PA-Modul wurde auf Basis
--    des ihb-Tickets https://tickets.ihb-eg.de/issues/4432 hinzugefuegt. Ab der fn2Module-Version vom 03.12.2018
--    (https://wiki.ihb-eg.de/doku.php/releasenotes/fn2module#section03122018) kann dieses "Flag" im PA-Modul auf dem Reiter "Person"
--    (dort rechts oben) geaendert werden.
-- c) Der Wert in dieser Spalte STUDENT.testPerson ist bei genau den Test-Studierenden 1, bei nicht Test-Studierenden ist der Wert etwas anderes als 1.
-- --------------------------------------------------------------------------------------------------------------------------------
/*
-- Hilfs-Select zum Herausfinden des "Standardwerts" in einer Spalte, also dem mit Abstand haeufigsten Wert:
-- Wieviel Prozent der Werte einer Spalte haben welchen Wert? Die Haeufigsten stehen oben.
-- Test-Studierende gehen nicht in das Ergebnis ein.
-- Einfach columnNameVar und tableNameVar austauschen, z.B. genen nachname und STUDENT.
SELECT
	columnNameVar AS wert,
	COUNT(*) AS anzahl,
	-- Anteil in Prozent, mit drei Nachkommastellen
	ROUND(
		(
			count(*) /
				(
					SELECT COUNT(*) FROM tableNameVar s1
					WHERE
						-- Ohne Test-Studierende (d.h. Student.testPerson darf nicht 1 sein)
						(COALESCE((SELECT s2.testPerson FROM flexnow.Student s2 WHERE s2.matrNr = s1.matrNr), 0) <> 1)
				)
		) * 100,
		3
	) AS anteil_prozent
FROM tableNameVar x
WHERE
	-- Ohne Test-Studierende (d.h. Student.testPerson darf nicht 1 sein)
	(COALESCE((SELECT s2.testPerson FROM flexnow.Student s2 WHERE s2.matrNr = x.matrNr), 0) <> 1)
GROUP BY columnNameVar
ORDER BY anzahl DESC;
*/
--
-- a) Pseudonymisierung der Tabelle flexnow.student, also im fn2Module-PA die Daten auf dem Reiter "Person"
--    Wenn das Statement mehrmals hintereinander ausgefuehrt wird (ohne dass sich durch eine andere Aktion die Daten der Tabelle Student geaendert haben),
--    bleiben die Daten so wie nach der ersten Ausfuehrung.
UPDATE flexnow.student stud
SET
	-- Der Wert sex = 0 steht fuer eine StudentIN, sex = 1 fuer einen Student.
	-- Bilden den neuen Vornamen: Nehme die ersten beiden Zeichen des richtigen Vornamen und haenge "Frau" bzw. "Mann", je nach Wert von sex. Beispiel: "PeMann".
	vorname    = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = stud.matrNr),
	dokVorname = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = stud.matrNr),
	-- Bilden den neuen Nachnamen: Nehme die ersten beiden Zeichen des richtigen Nachnamen und haenge die Matrikelnummer an. Beispiel: "Fa012345678".
	nachname = substr(nachname, 1, 2) || matrNrStr,
	-- nachnameascii enthaelt eigentlich den Nachnamen in Grossbuchstaben mit umgewandelten Sonderzeichen (aus z.B. "ü" wird "UE"). Spalte ist "not nullable".
	nachnameascii = substr(nachnameascii, 1, 2) || matrNrStr,
--
	-- Diese beiden Mail-Adressen gibt es wirklich
	email  = 'fn-dummy+test1@rub.de',
	email2 = 'fn-dummy+test2@rub.de',
	-- Alternativ: 'B.Sc.'
	akadGrad = 'Dr.',
	-- Die bei WiWi-Studierenden zu findenden Werte 'Dipl.-Ing.' und 'Dr.-Ing.' sind Akademische Grade, keine Namenszusaetze
	namensZusatz = 'von',
	-- 1. Nationalitaet [Dschibuti]. Spalte ist "not nullable".
	land = 'DJI',
	-- 2. Nationalitaet [China].
	zweitStaat = 'VRC',
--
	-- "g" fuer "Geburts...":
	gDatum = to_date('1980-01-01', 'YYYY-MM-DD'),
	gOrt = 'Hintertupfingen',
	gName = 'EinGeburtsname',
	-- [Dominica (Winward Inseln)]. Spalte ist "not nullable".
	gLand = 'WD',
--
	-- 1. Adresse: Adresse des Reichstags
	land1 = 'D', plz1 = 11011, ort1 = 'Berlin', strasse1 = 'Platz der Republik 1', vermieter1 = 'Vermieter A', telefon1 = '01234/56789',
	-- 2. Adresse: Adresse des Parlaments von Finnland
	land2 = 'FIN', plz2 = 00100, ort2 = 'Helsinki', strasse2 = 'Mannerheimintie 30', vermieter2 = 'Vermieter B', telefon2 = '04321/98765',
	-- 1, falls die erste Adresse die aktuelle Postanschrift ist, sonst 2; Spalte ist "not nullable".
	postAdresse = 1,
--
	-- Durch diese Quelle / Nutzer wurde der Datensatz erzeugt. 'Pseudonymisierer' steht fuer den Prozess, durch den die Studi-Daten pseudonymisiert werden, vor allem mit diesem SQL-Update-Ausdruck.
	nutzer = 'Pseudonymisierer',
	-- nutzername = ...,   Muss nicht veraendert werden, denn ist bei allen Studierenden: '108' || matrNrStr.
	passwort = 'ABCDE123',
--
--- Alle anderen Felder null/zero setzen oder auf ihren "Standardwert" (d.h. auf den am haeufigsten vorkommenden Wert):
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	univMatrNr = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	regId = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	personId = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	gebDat = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	homeVerzeichnis = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 1.
	univ = 1,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	bafoeg = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	hzb_art = null,
	-- "hzb" steht vermutlich fuer Hochschul-Zugangsberechtigung. War am 24.05.2019 fuer alle realen Studierenden -1.
	hzb_note = -1,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	orgEinheit = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	hsStatus = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	gastHoerer = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	studFach = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	studienFach = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	mentor = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	tutor = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	anderePerson = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	pFrage = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	pfAntwort = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	abschluss = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	mail_lock = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	note_erlaubt = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	nachteilAktiv = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	nachteilOffen = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	nachteilArt = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	plusKonzept = 0,
--
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	ndsNutzer = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	freischaltung = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	zertBeantragt = 0,
--
	-- War am 24.05.2019 fuer alle realen Studierenden 1.
	passwortAktiv = 1,
	-- War am 24.05.2019 fuer alle realen Studierenden 0.
	passwortAendern = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	passwortDatum = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 0. Spalte ist "not nullable".
	anzahlPasswort = 0,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	passwort_temp = null,
	-- War am 24.05.2019 fuer alle realen Studierenden 1.
	anzahlTan = 1,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	tansBekommen = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	tanDatum = null
WHERE
	-- Test-Studierenden (d.h. Student.testPerson darf nicht 1 sein) sollen nicht tangiert werden, also nicht pseudonymisiert werden.
	(COALESCE((SELECT s.testPerson FROM flexnow.Student s WHERE s.matrNr = stud.matrNr), 0) <> 1)
	-- Zum Testen dieses Statements an einem bestimmten Studierenden die folgende Zeile (also: AND matrNr = ...) de-kommentieren.
--	AND matrNr = 123
;
--
-- --------------------------------------------------------------------------------------------------------------------------------
-- b) Pseudonymisierung der Tabelle flexnow.zeugnisDruck_kopf, also im fn2Module-PA die pruefungsunabhaengigen Daten bei den "Zeugnissen" unter "Abschluesse"
UPDATE flexnow.zeugnisDruck_kopf zdk
SET
	-- Der Wert sex = 0 steht fuer eine StudentIN, sex = 1 fuer einen Student.
	-- Bilden den neuen Vornamen: Nehme die ersten beiden Zeichen des richtigen Vornamen und haenge "Frau" bzw. "Mann", je nach Wert von sex. Beispiel: "PeMann".
	vorname    = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = zdk.matrNr),
	dokVorname = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = zdk.matrNr),
	-- Bilden den neuen Nachnamen: Nehme die ersten beiden Zeichen des richtigen Nachnamen und haenge die Matrikelnummer an. Beispiel: "Fa012345678".
	nachname = substr(nachname, 1, 2) || matrikelNummer,
	-- Das "von" als Namenszusatz
	name = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = zdk.matrNr)
		|| ' von ' || substr(nachname, 1, 2) || matrikelNummer,
	-- Das ist der akademische Grad aus dem PO-Modul, der ist fuer jeden Studenten gesetzt.
	-- Das ist also nicht der bereits vom Studenten erworbene hoechste Abschluss und muss daher nicht pseudonymisiert werden:
--	gradKurz = 'Dr.',
--	gradLang = 'Dr.',
	-- Die bei WiWi-Studis zu findenden Werte 'Dipl.-Ing.' und 'Dr.-Ing.' sind Akademische Grade, keine Namenszusaetze
	namensZusatz = 'von',
--
	-- "g" fuer "Geburts...":
	geburtsDatum = '01.01.1980',
	geburtsOrt = 'Hintertupfingen',
	geburtsName = 'Eingeburtsname',
	geburtsInfo	= 'geboren am 01.01.1980 in Hintertupfingen',
	-- [Dominica (Winward Inseln)]. KFZ ist "WD".
	gLand = 'Dominica (Winward Inseln)'
WHERE
	-- Test-Studierenden (d.h. Student.testPerson darf nicht 1 sein) sollen nicht tangiert werden, also nicht pseudonymisiert werden.
	(COALESCE((SELECT s.testPerson FROM flexnow.Student s WHERE s.matrNr = zdk.matrNr), 0) <> 1)
	-- Zum Testen dieses Statements an einem bestimmten Studierenden die folgende Zeile (also: AND matrNr = ...) de-kommentieren.
--	AND matrNr = 123
;
--
-- --------------------------------------------------------------------------------------------------------------------------------
-- c) Pseudonymisierung der Tabelle flexnow.ss_sos, also die beim SOS-Import importierten Stammdaten der Studis
UPDATE flexnow.SS_SOS sos
SET
	-- Der Wert sex = 0 steht fuer eine StudentIN, sex = 1 fuer einen Student.
	-- Bilden den neuen Vornamen: Nehme die ersten beiden Zeichen des richtigen Vornamen und haenge "Frau" bzw. "Mann", je nach Wert von sex. Beispiel: "PeMann".
	vorname  = substr(vorname, 1, 2) || (SELECT (CASE WHEN s.sex = 0 THEN 'Frau' ELSE 'Mann' END) FROM flexnow.student s WHERE s.matrNr = sos.matrNr),
	-- Bilden den neuen Nachnamen: Nehme die ersten beiden Zeichen des richtigen Nachnamen und haenge die Matrikelnummer an. Beispiel: "Fa012345678".
	-- Fuelle dabei die Matrikelnummer vorn mit Nullen auf, so dass sie neuen Ziffern lang ist.
	nachname = substr(nachname, 1, 2) || LPAD(matrNr, 9, '0'),
--
	-- Diese Mail-Adresse gibt es wirklich
	email = 'fn-dummy+test1@rub.de',
	-- Alternativ: 'B.Sc.'
	akadGrad = 'Dr.',
	-- Die bei WiWi-Studis zu findenden Werte 'Dipl.-Ing.' und 'Dr.-Ing.' sind Akademische Grade, keine Namenszusaetze
	namensZusatz = 'von',
	-- 1. Nationalitaet [Dschibuti]
	nation = 'DJI',
--
	-- "g" fuer "Geburts...":
	gDatum = to_date('1980-01-01', 'YYYY-MM-DD'),
	gOrt = 'Hintertupfingen',
	gName = 'EinGeburtsname',
	-- [Dominica (Winward Inseln)]. War am 24.05.2019 fuer alle realen Studierenden null.
	gebLand = 'WD',
--
	-- 1. Adresse: Adresse des Reichstags
	plz1 = 11011, ort1 = 'Berlin', strasse1 = 'Platz der Republik 1', vermieter1 = 'Vermieter A', telefon1 = '01234/56789',
	-- 2. Adresse: Adresse des Parlaments von Finnland
	plz2 = 00100, ort2 = 'Helsinki', strasse2 = 'Mannerheimintie 30', vermieter2 = 'Vermieter B', telefon2 = '04321/98765',
--
	-- Bedeutung unklar. War am 24.05.2019 fuer alle realen Studierenden null. Ist in der Tabelle Student bei allen Studis: '108' || matrNrStr.
	nutzerName = null,
--
	-- Bedeutung unklar. War am 24.05.2019 fuer alle realen Studierenden 2.
	kzva = 2,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	kfz1 = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	kfz2 = null,
--
	-- "hzb" steht vermutlich fuer Hochschul-Zugangsberechtigung. War am 24.05.2019 fuer alle realen Studierenden null.
	hzb_art = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	hzb_note = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	bafoeg = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	uid_ = null,
	-- War am 24.05.2019 fuer alle realen Studierenden null.
	vert = null
WHERE
	-- Test-Studierenden (d.h. Student.testPerson darf nicht 1 sein) sollen nicht tangiert werden, also nicht pseudonymisiert werden.
	(COALESCE((SELECT s.testPerson FROM flexnow.Student s WHERE s.matrNr = sos.matrNr), 0) <> 1)
	-- Zum Testen dieses Statements an einem bestimmten Studierenden die folgende Zeile (also: AND matrNr = ...) de-kommentieren.
--	AND matrNr = 123
;
--
-- --------------------------------------------------------------------------------------------------------------------------------
-- d) Pseudonymisierung der Spalten bez und email der Tabelle fn2meta.PersonStub.
--    PersonStub enthaelt einen grossen Teil der eingeschriebenen Studierenden der RUB, auch etwa 18.000 Studierende,
--    die nicht in der Tabelle Student sind und auf die daher nicht FlexNow zugegriffen werden kann.
--
-- d1) Pseudonymisiert die bez, z.B. "Max, Mustermann" in "MaMann, Mu012345678".
--     Aber nur von Studierenden, also nicht von Pruefern oder Stellvertretern, und nur von Personen,
--     die es mit der gleichen ID (personId bzw. matrNr) in der Tabelle Student gibt.
--     Setzt voraus, dass die Tabelle Student bereits pseudonymisiert ist.
UPDATE fn2meta.PersonStub ps
SET
	bez = (SELECT s.vorname FROM flexnow.Student s WHERE ps.personId = s.matrNr)
--		|| ' von '
		|| ', '
		|| (SELECT s.nachname FROM flexnow.Student s WHERE ps.personId = s.matrNr),
	email  = 'fn-dummy+test1@rub.de'
WHERE
	-- Ohne Pruefer/Stellvertreter. Die haben in der RUB eine personId von etwa -540 bis 500.
	ps.personId > 10000
	AND exists (SELECT * FROM flexnow.Student s WHERE ps.personId = s.matrNr)
	-- Test-Studierenden (d.h. Student.testPerson darf nicht 1 sein) sollen nicht tangiert werden, also nicht pseudonymisiert werden.
	AND (COALESCE((SELECT s.testPerson FROM flexnow.Student s WHERE s.matrNr = ps.personId), 0) <> 1)
	-- Zum Testen dieses Statements an einem bestimmten Studierenden die folgende Zeile (also: AND matrNr = ...) de-kommentieren.
--	AND ps.personId = 014266089   -- bez: "Patrick Romann Nyadjou Kouatchou"
;
--
-- d2) Bei Studierenden, die nicht in der Tabelle Student sind, wird der Name durch einen fixen Text
--     und die personId/Matrikelnummer ersetzt. Denn diese Studierenden sollten staerker pseudonymisiert
--     werden als die "FlexNow-Studierenden".
UPDATE fn2meta.PersonStub ps
SET
	bez = ('[Bei Pseudonymisierung überschrieben; ' || LPAD(ps.personId, 9, '0') || ']'),
	email  = 'fn-dummy+test1@rub.de'
WHERE
	-- Ohne Pruefer/Stellvertreter. Die haben in der RUB eine personId von etwa -540 bis 500.
	ps.personId > 10000
	AND not exists (SELECT * FROM flexnow.Student s WHERE s.matrNr = ps.personId)
	-- Test-Studierenden sollen nicht tangiert werden, also nicht pseudonymisiert werden.
	-- Zum Testen dieses Statements an einem bestimmten Studierenden die folgende Zeile (also: AND matrNr = ...) de-kommentieren.
--	AND ps.personId = 99109548   -- bez: "Edin, Karasalihovic"
;
